"""Tests for the default scheduler"""

from collections.abc import Generator
from datetime import datetime, timedelta
from unittest.mock import MagicMock

import pytest
from madsci.client.resource_client import Resource
from madsci.common.types.action_types import ActionDefinition
from madsci.common.types.condition_types import (
    NoResourceInLocationCondition,
    ResourceInLocationCondition,
)
from madsci.common.types.location_types import Location, LocationDefinition
from madsci.common.types.node_types import Node, NodeInfo, NodeStatus
from madsci.common.types.resource_types import Slot
from madsci.common.types.step_types import Step
from madsci.common.types.workcell_types import WorkcellManagerDefinition
from madsci.common.types.workflow_types import (
    SchedulerMetadata,
    Workflow,
    WorkflowStatus,
)
from madsci.workcell_manager.schedulers.default_scheduler import Scheduler


@pytest.fixture
def mock_scheduler() -> Generator[Scheduler, None, None]:
    """Fixture to create a mock scheduler"""
    mock_workcell_definition = WorkcellManagerDefinition(
        name="test workcell",
        locations=[
            LocationDefinition(
                location_name="loc1",
                resource_id=None,
            ),
            LocationDefinition(
                location_name="loc2",
                resource_id=None,
            ),
        ],
        nodes={"test_node": "http://test_node"},
    )
    mock_state_handler = MagicMock()
    test_action = ActionDefinition(name="test_action", description="Test action")
    node_info = NodeInfo(
        node_name="test_node",
        module_name="test_module",
        actions={"test_action": test_action},
    )
    test_node = Node(
        node_url="http://test_node",
        status=NodeStatus(),
        info=node_info,
    )
    mock_state_handler.get_node.return_value = test_node
    mock_state_handler.get_nodes.return_value = {"test_node": test_node}

    # Mock the node_lock to return an unlocked lock by default
    mock_lock = MagicMock()
    mock_lock.locked.return_value = False
    mock_state_handler.node_lock.return_value = mock_lock

    scheduler = Scheduler(mock_workcell_definition, mock_state_handler)

    # Mock the LocationClient
    scheduler.location_client = MagicMock()
    scheduler.location_client.get_locations.return_value = [
        Location(
            name="loc1",
            resource_id=None,
        ),
        Location(
            name="loc2",
            resource_id=None,
        ),
    ]
    yield scheduler


@pytest.fixture
def workflows() -> list[Workflow]:
    """Fixture to create a list of workflows"""
    now = datetime.now()
    return [
        Workflow(
            name="wf1",
            submitted_time=now - timedelta(minutes=10),
            steps=[Step(name="step1", action="test_action", node="test_node")],
        ),
        Workflow(
            name="wf2",
            submitted_time=now - timedelta(minutes=5),
            steps=[Step(name="step2", action="test_action", node="test_node")],
        ),
    ]


def test_fifo_prioritization(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that workflows are prioritized based on FIFO from their submission dates"""
    result = mock_scheduler.run_iteration(workflows)
    assert (
        result[workflows[0].workflow_id].priority
        > result[workflows[1].workflow_id].priority
    )
    workflows_reversed = [workflows[1], workflows[0]]
    result = mock_scheduler.run_iteration(workflows_reversed)
    assert (
        result[workflows[0].workflow_id].priority
        > result[workflows[1].workflow_id].priority
    )


def test_condition_checking_no_resource_info_for_location(
    mock_scheduler: Scheduler,
) -> None:
    """Test that conditions are evaluated correctly when no resource information is available for a location"""
    workflows: list[Workflow] = [
        Workflow(
            name="wf1",
            submitted_time=datetime.now(),
            steps=[
                Step(
                    name="step1",
                    action="test_action",
                    node="test_node",
                    conditions=[
                        ResourceInLocationCondition(location_name="loc1"),
                        NoResourceInLocationCondition(location_name="loc2"),
                    ],
                )
            ],
        )
    ]
    mock_scheduler.resource_client = MagicMock()
    mock_scheduler.resource_client.get_resource.return_value = None

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    metadata = result[workflows[0].workflow_id]
    assert not metadata.ready_to_run
    assert len(metadata.reasons) > 0
    assert "does not have an attached container resource" in metadata.reasons[0]
    assert "does not have an attached container resource" in metadata.reasons[1]


def test_condition_checking_resource_presence(mock_scheduler: Scheduler) -> None:
    """Test that conditions are evaluated correctly when no resource information is available for a location"""
    workflows: list[Workflow] = [
        Workflow(
            name="wf1",
            submitted_time=datetime.now(),
            steps=[
                Step(
                    name="step1",
                    action="test_action",
                    node="test_node",
                    conditions=[
                        ResourceInLocationCondition(location_name="loc1"),
                        NoResourceInLocationCondition(location_name="loc2"),
                    ],
                )
            ],
        )
    ]
    mock_scheduler.resource_client = MagicMock()
    test_slot = Slot()
    mock_scheduler.resource_client.get_resource.return_value = test_slot
    mock_scheduler.location_client.get_locations.return_value[
        0
    ].resource_id = test_slot.resource_id
    mock_scheduler.location_client.get_locations.return_value[
        1
    ].resource_id = test_slot.resource_id

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    metadata = result[workflows[0].workflow_id]
    assert not metadata.ready_to_run
    assert len(metadata.reasons) > 0
    assert "does not contain" in metadata.reasons[0]

    mock_scheduler.resource_client.get_resource.return_value = Resource()

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    metadata = result[workflows[0].workflow_id]
    assert not metadata.ready_to_run
    assert len(metadata.reasons) > 0
    assert "is not a container." in metadata.reasons[0]
    assert "is not a container." in metadata.reasons[1]

    test_slot.children = [Resource()]
    mock_scheduler.resource_client.get_resource.return_value = test_slot

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    metadata = result[workflows[0].workflow_id]
    assert not metadata.ready_to_run
    assert len(metadata.reasons) > 0
    assert "contains" in metadata.reasons[0]


def test_workflow_paused(mock_scheduler: Scheduler, workflows: list[Workflow]) -> None:
    """Test that paused workflows are marked as not ready to run"""
    workflows[0].status.paused = True
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert "Workflow is paused" in result[workflows[0].workflow_id].reasons


def test_workflow_running(mock_scheduler: Scheduler, workflows: list[Workflow]) -> None:
    """Test that running workflows are marked as not ready to run"""
    workflows[0].status.running = True
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert "Workflow is already running" in result[workflows[0].workflow_id].reasons


def test_workflow_status_failed(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that workflows with an error status are not ready to run"""
    workflows[0].status.failed = True
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert "Workflow must be active" in result[workflows[0].workflow_id].reasons[0]


def test_workflow_status_completed(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that completed workflows are not ready to run"""
    workflows[0].status.completed = True
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert "Workflow must be active" in result[workflows[0].workflow_id].reasons[0]


def test_workflow_status_cancelled(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that cancelled workflows are not ready to run"""
    workflows[0].status.cancelled = True
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert "Workflow must be active" in result[workflows[0].workflow_id].reasons[0]


def test_workflow_status_queued(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that queued workflows are ready to run"""
    workflows[0].status = WorkflowStatus()
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert result[workflows[0].workflow_id].ready_to_run


def test_node_status_abnormal(
    mock_scheduler: Scheduler, workflows: list[Workflow]
) -> None:
    """Test that workflows are not ready to run if the node is in an error state"""
    test_action = ActionDefinition(name="test_action", description="Test action")
    mock_scheduler.state_handler.get_node.return_value = Node(
        node_url="http://test_node",
        status=NodeStatus(errored=True),
        info=NodeInfo(
            node_name="test_node",
            module_name="test_module",
            actions={"test_action": test_action},
        ),
    )
    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert (
        "Node test_node not ready: Node is in an error state"
        in result[workflows[0].workflow_id].reasons[0]
    )


def test_node_locked(mock_scheduler: Scheduler, workflows: list[Workflow]) -> None:
    """Test that workflows are not ready to run if the node is locked"""
    # Mock the node_lock to return a locked lock
    mock_lock = MagicMock()
    mock_lock.locked.return_value = True
    mock_scheduler.state_handler.node_lock.return_value = mock_lock

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert not result[workflows[0].workflow_id].ready_to_run
    assert (
        "Node test_node is locked by another action"
        in result[workflows[0].workflow_id].reasons
    )


def test_node_unlocked(mock_scheduler: Scheduler, workflows: list[Workflow]) -> None:
    """Test that workflows are ready to run if the node is not locked"""
    # Mock the node_lock to return an unlocked lock
    mock_lock = MagicMock()
    mock_lock.locked.return_value = False
    mock_scheduler.state_handler.node_lock.return_value = mock_lock

    result: dict[str, SchedulerMetadata] = mock_scheduler.run_iteration(workflows)
    assert result[workflows[0].workflow_id].ready_to_run


# TODO: Test Location Reservation
# TODO: Test Node Reservation



"""Unit tests for madsci.workcell_manager.workcell_actions module."""

from unittest.mock import MagicMock, Mock, patch

import pytest
from madsci.common.exceptions import LocationNotFoundError, WorkflowFailedError
from madsci.common.types.action_types import ActionStatus
from madsci.common.types.context_types import MadsciContext
from madsci.common.types.resource_types.server_types import ResourceHierarchy
from madsci.workcell_manager.workcell_actions import (
    _find_resource_and_transfer,
    _resolve_location_identifier,
    transfer,
    transfer_resource,
    wait,
)


class TestWaitAction:
    """Test cases for the wait action."""

    def test_wait_action_succeeds(self):
        """Test that wait action returns success after sleeping."""
        with patch("madsci.workcell_manager.workcell_actions.time.sleep") as mock_sleep:
            result = wait(5)

            mock_sleep.assert_called_with(5)
            assert result.status == ActionStatus.SUCCEEDED


class TestTransferAction:
    """Test cases for the simplified transfer action."""

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    def test_transfer_missing_location_server_url(self, mock_context):
        """Test that transfer fails when location server URL is not configured."""
        mock_context.return_value = MadsciContext(location_server_url=None)

        result = transfer("source_loc", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Location server URL not configured" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    def test_transfer_missing_workcell_server_url(self, mock_context):
        """Test that transfer fails when workcell server URL is not configured."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/", workcell_server_url=None
        )

        result = transfer("source_loc", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Workcell server URL not configured" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    def test_transfer_source_location_not_found(
        self, mock_location_client, mock_context
    ):
        """Test that transfer fails when source location is not found."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_client_instance = MagicMock()
        mock_location_client.return_value = mock_client_instance

        # Mock _resolve_location_identifier to raise exception for source
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = LocationNotFoundError(
                "Location 'nonexistent' not found by ID or name"
            )

            result = transfer("nonexistent", "dest_loc")

            assert result.status == ActionStatus.FAILED
            assert "Unexpected error in transfer" in result.errors[0].message
            assert "not found by ID or name" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    def test_transfer_no_path_exists(self, mock_location_client, mock_context):
        """Test that transfer fails when no transfer path exists."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_client_instance = MagicMock()
        mock_location_client.return_value = mock_client_instance

        # Mock successful location resolution
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = ["source_id", "dest_id"]

            # Mock plan_transfer to raise exception (no path)
            mock_client_instance.plan_transfer.side_effect = Exception(
                "No transfer path exists between source_loc and dest_loc"
            )

            result = transfer("source_loc", "dest_loc")

            assert result.status == ActionStatus.FAILED
            assert "Unexpected error in transfer" in result.errors[0].message
            assert "No transfer path exists" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.WorkcellClient")
    def test_transfer_success(
        self, mock_workcell_client, mock_location_client, mock_context
    ):
        """Test successful transfer action."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock workcell client
        mock_workcell_instance = MagicMock()
        mock_workcell_client.return_value = mock_workcell_instance

        # Mock successful location resolution
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = ["source_id", "dest_id"]

            # Mock successful transfer planning
            mock_workflow_def = {"name": "test_transfer", "steps": []}
            mock_location_instance.plan_transfer.return_value = mock_workflow_def

            # Mock successful workflow execution
            mock_workflow = MagicMock()
            mock_workflow.workflow_id = "test_workflow_123"
            mock_workflow.status.completed = True
            mock_workflow.status.workflow_runtime = 5.0
            mock_workflow.duration_seconds = 5.0
            mock_workcell_instance.start_workflow.return_value = mock_workflow

            result = transfer("source_loc", "dest_loc")

            assert result.status == ActionStatus.SUCCEEDED
            assert "Transfer completed" in result.json_result["message"]
            assert result.json_result["source_location_id"] == "source_id"
            assert result.json_result["target_location_id"] == "dest_id"
            assert result.json_result["workflow_id"] == "test_workflow_123"

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.WorkcellClient")
    def test_transfer_workflow_failed(
        self, mock_workcell_client, mock_location_client, mock_context
    ):
        """Test transfer action when workflow execution fails."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock workcell client
        mock_workcell_instance = MagicMock()
        mock_workcell_client.return_value = mock_workcell_instance

        # Mock successful location resolution
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = ["source_id", "dest_id"]

            # Mock successful transfer planning
            mock_workflow_def = {"name": "test_transfer", "steps": []}
            mock_location_instance.plan_transfer.return_value = mock_workflow_def

            # Mock failed workflow execution
            mock_workflow = MagicMock()
            mock_workflow.workflow_id = "test_workflow_123"
            mock_workflow.status.completed = False
            mock_workflow.status.failed = True
            mock_workflow.status.description = "Robot arm malfunction"
            mock_workflow.status.current_step_index = 2
            mock_workcell_instance.start_workflow.return_value = mock_workflow

            result = transfer("source_loc", "dest_loc")

            assert result.status == ActionStatus.FAILED
            assert "Transfer workflow failed at step 2" in result.errors[0].message
            assert "Robot arm malfunction" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.WorkcellClient")
    def test_transfer_async_execution(
        self, mock_workcell_client, mock_location_client, mock_context
    ):
        """Test transfer action with async execution (await_completion=False)."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock workcell client
        mock_workcell_instance = MagicMock()
        mock_workcell_client.return_value = mock_workcell_instance

        # Mock successful location resolution
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = ["source_id", "dest_id"]

            # Mock successful transfer planning
            mock_workflow_def = {"name": "test_transfer", "steps": []}
            mock_location_instance.plan_transfer.return_value = mock_workflow_def

            # Mock workflow enqueueing (not completed since await_completion=False)
            mock_workflow = MagicMock()
            mock_workflow.workflow_id = "test_workflow_123"
            mock_workcell_instance.start_workflow.return_value = mock_workflow

            result = transfer("source_loc", "dest_loc", await_completion=False)

            assert result.status == ActionStatus.SUCCEEDED
            assert "Transfer workflow enqueued" in result.json_result["message"]
            assert result.json_result["workflow_id"] == "test_workflow_123"

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    def test_transfer_exception_handling(self, mock_location_client, mock_context):
        """Test that transfer handles unexpected exceptions gracefully."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client to raise exception
        mock_location_client.side_effect = Exception("Connection error")

        result = transfer("source_loc", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Unexpected error in transfer" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.WorkcellClient")
    def test_transfer_workflow_failed_exception(
        self, mock_workcell_client, mock_location_client, mock_context
    ):
        """Test transfer action when WorkflowFailedError is raised."""

        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/",
            workcell_server_url="http://localhost:8005/",
        )

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock workcell client to raise WorkflowFailedError
        mock_workcell_instance = MagicMock()
        mock_workcell_client.return_value = mock_workcell_instance

        # Mock successful location resolution
        with patch(
            "madsci.workcell_manager.workcell_actions._resolve_location_identifier"
        ) as mock_resolve:
            mock_resolve.side_effect = ["source_id", "dest_id"]

            # Mock successful transfer planning
            mock_workflow_def = {"name": "test_transfer", "steps": []}
            mock_location_instance.plan_transfer.return_value = mock_workflow_def

            # Mock WorkflowFailedError
            mock_workcell_instance.start_workflow.side_effect = WorkflowFailedError(
                "Step execution failed"
            )

            result = transfer("source_loc", "dest_loc")

            assert result.status == ActionStatus.FAILED
            assert (
                "Transfer workflow failed during execution" in result.errors[0].message
            )
            assert "Step execution failed" in result.errors[0].message


class TestResolveLocationIdentifier:
    """Test cases for the _resolve_location_identifier helper function."""

    def test_resolve_by_id_success(self):
        """Test successful location resolution by ID."""
        mock_client = MagicMock()
        mock_location = Mock()
        mock_location.location_id = "test_id"
        mock_client.get_location.return_value = mock_location

        result = _resolve_location_identifier("test_id", mock_client)

        assert result == "test_id"
        mock_client.get_location.assert_called_once_with("test_id")

    def test_resolve_by_name_success(self):
        """Test successful location resolution by name."""
        mock_client = MagicMock()

        # Mock get_location to fail (not found by ID)
        mock_client.get_location.side_effect = Exception("Not found")

        # Mock get_location_by_name to return matching location
        mock_location = Mock()
        mock_location.location_id = "resolved_id"
        mock_location.location_name = "test_name"
        mock_client.get_location_by_name.return_value = mock_location

        result = _resolve_location_identifier("test_name", mock_client)

        assert result == "resolved_id"

    def test_resolve_not_found(self):
        """Test location resolution failure when location doesn't exist."""
        mock_client = MagicMock()

        # Mock get_location to fail
        mock_client.get_location.side_effect = Exception("Not found")

        # Mock get_location_by_name to fail
        mock_client.get_location_by_name.side_effect = Exception("Not found")

        with pytest.raises(LocationNotFoundError) as exc_info:
            _resolve_location_identifier("nonexistent", mock_client)

        assert "not found by ID or name" in str(exc_info.value)

    def test_resolve_client_error(self):
        """Test location resolution when client throws unexpected error."""
        mock_client = MagicMock()

        # Mock both methods to fail
        mock_client.get_location.side_effect = Exception("Connection error")
        mock_client.get_location_by_name.side_effect = Exception("Connection error")

        with pytest.raises(LocationNotFoundError) as exc_info:
            _resolve_location_identifier("test", mock_client)

        # The function catches exceptions internally and raises "not found" instead of connection error
        assert "not found by ID or name" in str(exc_info.value)


class TestTransferResourceAction:
    """Test cases for the transfer_resource action."""

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    def test_transfer_resource_missing_resource_server_url(self, mock_context):
        """Test that transfer_resource fails when resource server URL is not configured."""
        mock_context.return_value = MadsciContext(
            location_server_url="http://localhost:8006/", resource_server_url=None
        )

        result = transfer_resource("resource_123", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Resource server URL not configured" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    def test_transfer_resource_missing_location_server_url(self, mock_context):
        """Test that transfer_resource fails when location server URL is not configured."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/", location_server_url=None
        )

        result = transfer_resource("resource_123", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Location server URL not configured" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions._find_resource_and_transfer")
    def test_transfer_resource_invalid_resource(
        self, mock_find_and_transfer, mock_context
    ):
        """Test that transfer_resource fails when resource doesn't exist."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/",
            location_server_url="http://localhost:8006/",
        )

        # Mock helper function to return error message
        mock_find_and_transfer.return_value = (
            "Failed to verify resource 'invalid_resource': Resource not found"
        )

        result = transfer_resource("invalid_resource", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert "Failed to verify resource" in result.errors[0].message

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions._find_resource_and_transfer")
    def test_transfer_resource_not_found_at_location(
        self, mock_find_and_transfer, mock_context
    ):
        """Test that transfer_resource fails when resource is not found at any location."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/",
            location_server_url="http://localhost:8006/",
        )

        # Mock helper function to return error message
        mock_find_and_transfer.return_value = (
            "Resource 'resource_123' not found at any location"
        )

        result = transfer_resource("resource_123", "dest_loc")

        assert result.status == ActionStatus.FAILED
        assert (
            "Resource 'resource_123' not found at any location"
            in result.errors[0].message
        )

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions._find_resource_and_transfer")
    def test_transfer_resource_success(self, mock_find_and_transfer, mock_context):
        """Test successful transfer_resource action."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/",
            location_server_url="http://localhost:8006/",
        )

        # Mock successful transfer result
        mock_transfer_result = MagicMock()
        mock_transfer_result.status = ActionStatus.SUCCEEDED
        mock_transfer_result.json_result = {
            "message": "Transfer completed",
            "source_location_id": "source_location_id",
            "target_location_id": "dest_location_id",
            "workflow_id": "workflow_123",
        }
        mock_find_and_transfer.return_value = mock_transfer_result

        result = transfer_resource("resource_123", "dest_loc")

        assert result.status == ActionStatus.SUCCEEDED
        assert result.json_result["message"] == "Transfer completed"
        mock_find_and_transfer.assert_called_once_with(
            "resource_123", "dest_loc", True, mock_context.return_value
        )

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions._find_resource_and_transfer")
    def test_transfer_resource_transfer_failed(
        self, mock_find_and_transfer, mock_context
    ):
        """Test transfer_resource when underlying transfer fails."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/",
            location_server_url="http://localhost:8006/",
        )

        # Mock failed transfer result
        mock_transfer_result = MagicMock()
        mock_transfer_result.status = ActionStatus.FAILED
        mock_find_and_transfer.return_value = mock_transfer_result

        result = transfer_resource("resource_123", "dest_loc")

        assert result.status == ActionStatus.FAILED
        mock_find_and_transfer.assert_called_once_with(
            "resource_123", "dest_loc", True, mock_context.return_value
        )

    @patch("madsci.workcell_manager.workcell_actions.get_current_madsci_context")
    @patch("madsci.workcell_manager.workcell_actions._find_resource_and_transfer")
    def test_transfer_resource_async_execution(
        self, mock_find_and_transfer, mock_context
    ):
        """Test transfer_resource with async execution (await_completion=False)."""
        mock_context.return_value = MadsciContext(
            resource_server_url="http://localhost:8003/",
            location_server_url="http://localhost:8006/",
        )

        # Mock successful async transfer result
        mock_transfer_result = MagicMock()
        mock_transfer_result.status = ActionStatus.SUCCEEDED
        mock_transfer_result.json_result = {
            "message": "Transfer workflow enqueued",
            "workflow_id": "workflow_123",
        }
        mock_find_and_transfer.return_value = mock_transfer_result

        result = transfer_resource("resource_123", "dest_loc", await_completion=False)

        assert result.status == ActionStatus.SUCCEEDED
        assert result.json_result["message"] == "Transfer workflow enqueued"
        mock_find_and_transfer.assert_called_once_with(
            "resource_123", "dest_loc", False, mock_context.return_value
        )


class TestFindResourceAndTransferHelper:
    """Test cases for the _find_resource_and_transfer helper function."""

    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.ResourceClient")
    @patch("madsci.workcell_manager.workcell_actions.transfer")
    def test_find_resource_and_transfer_success(
        self, mock_transfer, mock_resource_client, mock_location_client
    ):
        """Test successful resource finding and transfer."""
        # Create mock context
        mock_context = MagicMock()
        mock_context.location_server_url = "http://localhost:8006/"
        mock_context.resource_server_url = "http://localhost:8003/"

        # Mock resource client
        mock_resource_instance = MagicMock()
        mock_resource_client.return_value = mock_resource_instance
        mock_resource_instance.get_resource.return_value = (
            MagicMock()
        )  # Resource exists

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock locations
        mock_location1 = Mock()
        mock_location1.location_id = "loc_1"
        mock_location2 = Mock()
        mock_location2.location_id = "loc_2"
        mock_location_instance.get_locations.return_value = [
            mock_location1,
            mock_location2,
        ]

        # Mock location resources - first location doesn't have resource, second does
        empty_hierarchy = ResourceHierarchy(
            ancestor_ids=[], resource_id="", descendant_ids={}
        )
        found_hierarchy = ResourceHierarchy(
            ancestor_ids=[], resource_id="resource_123", descendant_ids={}
        )
        mock_location_instance.get_location_resources.side_effect = [
            empty_hierarchy,
            found_hierarchy,
        ]

        # Mock successful transfer
        mock_transfer_result = MagicMock()
        mock_transfer_result.status = ActionStatus.SUCCEEDED
        mock_transfer.return_value = mock_transfer_result

        result = _find_resource_and_transfer(
            "resource_123", "dest_loc", True, mock_context
        )

        assert result == mock_transfer_result
        mock_transfer.assert_called_once_with(
            source="loc_2", target="dest_loc", await_completion=True
        )

    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.ResourceClient")
    def test_find_resource_and_transfer_resource_not_found(
        self, mock_resource_client, mock_location_client
    ):
        """Test when resource doesn't exist."""
        # Create mock context
        mock_context = MagicMock()
        mock_context.location_server_url = "http://localhost:8006/"
        mock_context.resource_server_url = "http://localhost:8003/"
        _ = mock_location_client

        # Mock resource client to raise exception
        mock_resource_instance = MagicMock()
        mock_resource_client.return_value = mock_resource_instance
        mock_resource_instance.get_resource.side_effect = Exception(
            "Resource not found"
        )

        result = _find_resource_and_transfer(
            "invalid_resource", "dest_loc", True, mock_context
        )

        assert isinstance(result, str)
        assert "Failed to verify resource" in result

    @patch("madsci.workcell_manager.workcell_actions.LocationClient")
    @patch("madsci.workcell_manager.workcell_actions.ResourceClient")
    def test_find_resource_and_transfer_location_not_found(
        self, mock_resource_client, mock_location_client
    ):
        """Test when resource is not found at any location."""
        # Create mock context
        mock_context = MagicMock()
        mock_context.location_server_url = "http://localhost:8006/"
        mock_context.resource_server_url = "http://localhost:8003/"

        # Mock resource client
        mock_resource_instance = MagicMock()
        mock_resource_client.return_value = mock_resource_instance
        mock_resource_instance.get_resource.return_value = (
            MagicMock()
        )  # Resource exists

        # Mock location client
        mock_location_instance = MagicMock()
        mock_location_client.return_value = mock_location_instance

        # Mock locations
        mock_location1 = Mock()
        mock_location1.location_id = "loc_1"
        mock_location_instance.get_locations.return_value = [mock_location1]

        # Mock location resources - location doesn't have the resource
        empty_hierarchy = ResourceHierarchy(
            ancestor_ids=[], resource_id="", descendant_ids={}
        )
        mock_location_instance.get_location_resources.return_value = empty_hierarchy

        result = _find_resource_and_transfer(
            "resource_123", "dest_loc", True, mock_context
        )

        assert isinstance(result, str)
        assert "not found at any location" in result



"""
Engine Class and associated helpers and data
"""

import concurrent
import importlib
import time
import traceback
from datetime import datetime
from typing import Optional, Union

from madsci.client.data_client import DataClient
from madsci.client.event_client import EventClient
from madsci.client.location_client import LocationClient
from madsci.client.node.abstract_node_client import AbstractNodeClient
from madsci.client.resource_client import ResourceClient
from madsci.common.nodes import check_node_capability
from madsci.common.ownership import ownership_context
from madsci.common.types.action_types import (
    ActionDatapoints,
    ActionFiles,
    ActionRequest,
    ActionResult,
    ActionStatus,
)
from madsci.common.types.base_types import Error
from madsci.common.types.datapoint_types import (
    DataPoint,
    DataPointTypeEnum,
    FileDataPoint,
    ValueDataPoint,
)
from madsci.common.types.event_types import Event, EventType
from madsci.common.types.node_types import Node, NodeStatus
from madsci.common.types.parameter_types import (
    ParameterFeedForwardFile,
    ParameterFeedForwardJson,
)
from madsci.common.types.step_types import Step
from madsci.common.types.workflow_types import Workflow
from madsci.common.utils import threaded_daemon
from madsci.workcell_manager.state_handler import WorkcellStateHandler
from madsci.workcell_manager.workcell_actions import workcell_action_dict
from madsci.workcell_manager.workcell_utils import (
    find_node_client,
)
from madsci.workcell_manager.workflow_utils import (
    cancel_active_workflows,
    prepare_workflow_step,
)


class Engine:
    """
    Handles scheduling workflows and executing steps on the workcell.
    Pops incoming workflows off a redis-based queue and executes them.
    """

    def __init__(
        self, state_handler: WorkcellStateHandler, data_client: DataClient
    ) -> None:
        """Initialize the scheduler."""
        self.state_handler = state_handler
        self.workcell_definition = state_handler.get_workcell_definition()
        self.workcell_settings = self.state_handler.workcell_settings
        self.logger = EventClient(name=f"workcell.{self.workcell_definition.name}")
        cancel_active_workflows(state_handler)
        scheduler_module = importlib.import_module(self.workcell_settings.scheduler)
        self.scheduler = scheduler_module.Scheduler(
            self.workcell_definition, self.state_handler
        )
        self.data_client = data_client
        self.resource_client = ResourceClient()
        self.location_client = LocationClient()
        with state_handler.wc_state_lock():
            state_handler.initialize_workcell_state()
        time.sleep(self.workcell_settings.cold_start_delay)
        self.logger.info("Engine initialized, waiting for workflows...")

    @threaded_daemon
    def spin(self) -> None:
        """
        Continuously loop, updating node states every Config.update_interval seconds.
        If the state of the workcell has changed, update the active modules and run the scheduler.
        """
        self.update_active_nodes(self.state_handler, update_info=True)
        node_tick = time.time()
        info_tick = time.time()
        scheduler_tick = time.time()
        while True and not self.state_handler.shutdown:
            try:
                self.workcell_definition = self.state_handler.get_workcell_definition()

                # Check if it's time to update node info (less frequent)
                should_update_info = (
                    time.time() - info_tick
                    > self.workcell_settings.node_info_update_interval
                )

                if (
                    time.time() - node_tick
                    > self.workcell_settings.node_update_interval
                ):
                    self.update_active_nodes(
                        self.state_handler, update_info=should_update_info
                    )
                    node_tick = time.time()
                    if should_update_info:
                        info_tick = time.time()
                if (
                    time.time() - scheduler_tick
                    > self.workcell_settings.scheduler_update_interval
                ):
                    with self.state_handler.wc_state_lock():
                        self.state_handler.update_workflow_queue()
                        self.state_handler.archive_terminal_workflows()
                        workflows = self.state_handler.get_workflow_queue()
                        workflow_definition_metadata_map = self.scheduler.run_iteration(
                            workflows=workflows
                        )
                        for workflow in workflows:
                            if workflow.workflow_id in workflow_definition_metadata_map:
                                workflow.scheduler_metadata = (
                                    workflow_definition_metadata_map[
                                        workflow.workflow_id
                                    ]
                                )
                                self.state_handler.set_active_workflow(
                                    workflow, mark_state_changed=False
                                )
                            else:
                                workflow.scheduler_metadata.ready_to_run = False
                                self.state_handler.set_active_workflow(
                                    workflow, mark_state_changed=False
                                )
                    if self.state_handler.get_workcell_status().ok:
                        self.run_next_step()
                        scheduler_tick = time.time()
            except Exception as e:
                self.logger.error(e)
                self.logger.warning(
                    f"Error in engine loop, waiting {10 * self.workcell_settings.node_update_interval} seconds before trying again."
                )
                with self.state_handler.wc_state_lock():
                    workcell_status = self.state_handler.get_workcell_status()
                    workcell_status.errored = True
                    workcell_status.errors.append(Error.from_exception(e))
                    self.state_handler.set_workcell_status(workcell_status)
                time.sleep(self.workcell_settings.node_update_interval)

    def run_next_step(self, await_step_completion: bool = False) -> Optional[Workflow]:
        """Runs the next step in the workflow with the highest priority. Returns information about the workflow it ran, if any."""
        next_wf = None
        with self.state_handler.wc_state_lock():
            workflows = self.state_handler.get_workflow_queue()
            ready_workflows = filter(
                lambda wf: wf.scheduler_metadata.ready_to_run, workflows
            )
            sorted_ready_workflows = sorted(
                ready_workflows,
                key=lambda wf: wf.scheduler_metadata.priority,
                reverse=True,
            )
            while len(sorted_ready_workflows) > 0:
                next_wf = sorted_ready_workflows[0]
                # * Check if the workflow is already complete
                if next_wf.status.current_step_index >= len(next_wf.steps):
                    self.logger.warning(
                        f"Workflow {next_wf.workflow_id} has no more steps, marking as completed"
                    )
                    next_wf.status.completed = True
                    self.state_handler.set_active_workflow(next_wf)
                    self._log_workflow_completion(next_wf, "completed")
                    sorted_ready_workflows.pop(0)
                    next_wf = None
                    continue
                next_wf = sorted_ready_workflows[0]
                next_wf.status.running = True
                next_wf.status.has_started = True
                if next_wf.status.current_step_index == 0:
                    next_wf.start_time = datetime.now()
                self.state_handler.set_active_workflow(next_wf)
                break
            else:
                self.logger.info("No workflows ready to run")
        if next_wf:
            thread = self.run_step(next_wf.workflow_id)
            if await_step_completion:
                thread.join()
        return next_wf

    @threaded_daemon
    def run_step(self, workflow_id: str) -> None:
        """Run a step in a standalone thread, updating the workflow as needed"""
        try:
            # * Prepare the step
            wf = self.state_handler.get_active_workflow(workflow_id)
            step = wf.steps[wf.status.current_step_index]
            step = prepare_workflow_step(
                step=step,
                workcell=self.workcell_definition,
                state_handler=self.state_handler,
                workflow=wf,
                data_client=self.data_client,
                location_client=self.location_client,
            )
            step.start_time = datetime.now()
            self.logger.info(f"Running step {step.step_id} in workflow {workflow_id}")
            if step.node is None:
                step = self.run_workcell_action(step)
            else:
                node = self.state_handler.get_node(step.node)
                client = find_node_client(node.node_url)
                wf = self.update_step(wf, step)

                # * Send the action request
                response = None

                # Merge with step.args
                args = {**step.args, **step.locations}
                request = ActionRequest(
                    action_name=step.action,
                    args=args,
                    files=step.file_paths,
                )
                action_id = request.action_id

                # Acquire lock on the node for the entire action duration
                node_lock = self.state_handler.node_lock(step.node)
                if not node_lock.acquire():
                    raise RuntimeError(f"Failed to acquire lock on node {step.node}")

                try:
                    self.logger.log_info(
                        f"Acquired lock on Node {step.node} for Action {action_id} in Step {step.step_id} of Workflow {workflow_id}"
                    )

                    try:
                        response = client.send_action(request, await_result=False)
                    except Exception as e:
                        self.logger.error(
                            f"Sending Action Request {action_id} for step {step.step_id} triggered exception: {e!s}"
                        )
                        if response is None:
                            # Create a running response so monitor_action_progress can try get_action_result
                            # as a fallback in case the action was actually created but the response was lost
                            response = request.running(errors=[Error.from_exception(e)])
                        else:
                            response.errors.append(Error.from_exception(e))

                    response = self.handle_response(wf, step, response)
                    action_id = response.action_id

                    # * Periodically query the action status until complete, updating the workflow as needed
                    # * If the node or client supports get_action_result, query the action result
                    node_lock.release()
                    self.monitor_action_progress(
                        wf, step, node, client, response, request, action_id
                    )
                finally:
                    self.logger.log_info(
                        f"Released lock on Node {step.node} for Action {action_id} in Step {step.step_id} of Workflow {workflow_id}"
                    )
                    if node_lock.locked():
                        node_lock.release()
                # * Finalize the step
            self.finalize_step(workflow_id, step)
            self.logger.info(f"Completed step {step.step_id} in workflow {workflow_id}")
            self.logger.debug(self.state_handler.get_workflow(workflow_id))
        except Exception as e:
            self.logger.error(
                f"Running step in workflow {workflow_id} triggered unhandled exception: {traceback.format_exc()}"
            )
            step.result = ActionResult(
                status=ActionStatus.FAILED,
                errors=Error.from_exception(e),
            )
            wf = self.update_step(wf, step)
            self.finalize_step(workflow_id, step)

    def run_workcell_action(self, step: Step) -> Step:
        """Runs one of the built-in workcell actions"""
        action_callable = workcell_action_dict[step.action]
        step.result = action_callable(**step.args)
        step.status = step.result.status
        return step

    def monitor_action_progress(
        self,
        wf: Workflow,
        step: Step,
        node: Node,
        client: AbstractNodeClient,
        response: ActionResult,
        request: ActionRequest,
        action_id: str,
    ) -> None:
        """Monitor the progress of the action, querying the action result until it is terminal"""
        interval = 1.0
        retry_count = 0
        while not response.status.is_terminal:
            if not check_node_capability(
                node_info=node.info, client=client, capability="get_action_result"
            ) and not check_node_capability(
                node_info=node.info, client=client, capability="get_action_status"
            ):
                self.logger.warning(
                    f"While running Step {step.step_id} of workflow {wf.workflow_id}, send_action returned a non-terminal response {response}. However, node {step.node} does not support querying an action result."
                )
                break
            try:
                time.sleep(interval)  # * Exponential backoff with cap
                interval = interval * 1.5 if interval < 10.0 else 10.0
                response = client.get_action_result(action_id)
                self.handle_response(wf, step, response)
                if (
                    response.status.is_terminal
                    or response.status == ActionStatus.UNKNOWN
                ):
                    # * If the action is terminal or unknown, break out of the loop
                    # * If the action is unknown, that means the node does not have a record of the action
                    break
            except Exception as e:
                self.logger.error(
                    f"Querying action {action_id} for step {step.step_id} resulted in exception: {e!s}"
                )
                if response is None:
                    response = request.unknown(errors=[Error.from_exception(e)])
                else:
                    response.errors.append(Error.from_exception(e))
                self.handle_response(wf, step, response)
                if retry_count >= self.workcell_settings.get_action_result_retries:
                    self.logger.error(
                        f"Exceeded maximum number of retries for querying action {action_id} for step {step.step_id}"
                    )
                    # Set status to UNKNOWN after exhausting retries
                    response = request.unknown(errors=response.errors)
                    self.handle_response(wf, step, response)
                    break
                retry_count += 1

    def update_param_value_from_datapoint(
        self,
        wf: Workflow,
        datapoint: DataPoint,
        parameter: Union[ParameterFeedForwardJson, ParameterFeedForwardFile],
    ) -> Workflow:
        """Updates the parameters in a workflow"""

        if datapoint.data_type == DataPointTypeEnum.JSON:
            wf.parameter_values[parameter.key] = datapoint.value
        elif datapoint.data_type in {
            DataPointTypeEnum.FILE,
            DataPointTypeEnum.OBJECT_STORAGE,
        }:
            wf.file_input_ids[parameter.key] = datapoint.datapoint_id
        return wf

    def finalize_step(self, workflow_id: str, step: Step) -> None:
        """Finalize the step, updating the workflow based on the results (setting status, updating index, etc.)"""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_active_workflow(workflow_id)
            step.end_time = datetime.now()
            wf.steps[wf.status.current_step_index] = step
            wf = self._feed_data_forward(step, wf)
            wf.status.running = False

            if step.status == ActionStatus.SUCCEEDED:
                new_index = wf.status.current_step_index + 1
                if new_index >= len(wf.steps):
                    wf.status.completed = True
                    wf.end_time = datetime.now()
                else:
                    wf.status.current_step_index = new_index
            elif step.status == ActionStatus.FAILED:
                wf.status.failed = True
                wf.end_time = datetime.now()
            elif step.status == ActionStatus.CANCELLED:
                wf.status.cancelled = True
                wf.end_time = datetime.now()
            elif step.status == ActionStatus.NOT_READY:
                pass
            elif step.status == ActionStatus.UNKNOWN:
                self.logger.error(
                    f"Step {step.step_id} in workflow {workflow_id} ended with unknown status"
                )
                wf.status.failed = True
                wf.end_time = datetime.now()
            else:
                self.logger.error(
                    f"Step {step.step_id} in workflow {workflow_id} ended with unexpected status {step.status}"
                )
                wf.status.failed = True
                wf.end_time = datetime.now()
            self.state_handler.set_active_workflow(wf)

            if wf.status.terminal:
                self.logger.log_info(str(wf))
                self._log_completion_event(wf)

    def _feed_data_forward(self, step: Step, wf: Workflow) -> Workflow:
        """Feed data forward from the completed step to the workflow parameters"""
        for param in wf.parameters.feed_forward:
            if (isinstance(param.step, str) and param.step == step.key) or (
                isinstance(param.step, int)
                and wf.status.current_step_index == param.step
            ):
                if step.result.datapoints:
                    self.logger.log_warning(
                        event=Event(event_data=str(step.result.datapoints))
                    )
                    datapoint_ids = step.result.datapoints.model_dump()
                    # Fetch actual DataPoint objects from data manager using the IDs
                    datapoint_dict = {}
                    for key, datapoint_id in datapoint_ids.items():
                        datapoint = self.data_client.get_datapoint(datapoint_id)
                        datapoint_dict[key] = datapoint
                else:
                    raise ValueError(
                        f"Feed-forward parameter {param.key} specified step {param.step} but step has no datapoints"
                    )
                if param.label is None:
                    self.logger.log_warning(event=Event(event_data=str(datapoint_dict)))
                    if len(datapoint_dict) == 1:
                        datapoint = next(iter(datapoint_dict.values()))
                        wf = self.update_param_value_from_datapoint(
                            wf, datapoint, param
                        )
                    else:
                        raise ValueError(
                            f"Ambiguous feed-forward parameter {param.key}: multiple possible matching labels: {step.result.datapoints.model_dump().keys()}"
                        )
                elif param.label in datapoint_dict:
                    datapoint = datapoint_dict[param.label]
                    wf = self.update_param_value_from_datapoint(wf, datapoint, param)
                elif param.step is not None:
                    raise ValueError(
                        f"Feed-forward parameter {param.key}'s specified label {param.label} not found in step result datapoints: {step.result.datapoints.model_dump().keys()}"
                    )
        return wf

    def _log_completion_event(self, workflow: Workflow) -> None:
        """Log the completion event and info message."""
        try:
            event_data = workflow.model_dump(mode="json")
            self.logger.log(
                Event(event_type=EventType.WORKFLOW_COMPLETE, event_data=event_data)
            )

            duration_text = (
                f"Duration: {event_data['duration_seconds']:.1f}s"
                if event_data["duration_seconds"]
                else "Duration: Unknown"
            )

            self.logger.info(
                f"Logged workflow completion: {workflow.name} ({workflow.workflow_id[-8:]}) - "
                f"Status: {event_data['status']}, Author: {(event_data.get('definition_metadata') or {}).get('author') or 'Unknown'}, "
                f"{duration_text}"
            )
        except Exception as e:
            self.logger.error(
                f"Error logging workflow completion event for workflow {workflow.workflow_id}: {e!s}\n{traceback.format_exc()}"
            )

    def update_step(self, wf: Workflow, step: Step) -> None:
        """Update the step in the workflow"""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_workflow(wf.workflow_id)
            wf.steps[wf.status.current_step_index] = step
            self.state_handler.set_active_workflow(wf)
        return wf

    def handle_response(
        self, wf: Workflow, step: Step, response: ActionResult
    ) -> Optional[ActionResult]:
        """Handle the response from the node"""
        response = self.handle_data_and_files(step, wf, response)
        step.status = response.status
        if response.status == ActionStatus.UNKNOWN:
            response.errors.append(
                Error(
                    message="Node returned 'unknown' action status for running action.",
                    error_type="NodeReturnedUnknown",
                )
            )
        step.result = response
        if (
            len(step.history) == 0
            or step.history[-1].action_id != response.action_id
            or step.history[-1].status != response.status
        ):
            # * Don't append redundant status updates
            step.history.append(response)
        step.last_status_update = datetime.now()
        wf = self.update_step(wf, step)
        return response

    def handle_data_and_files(
        self, step: Step, wf: Workflow, response: ActionResult
    ) -> ActionResult:
        """Upload non-datapoint results as datapoints and consolidate all datapoint IDs.

        This method ensures that all results (JSON data, files) are stored as datapoints
        in the data manager, following the principle of getting data into the data manager ASAP.
        The response datapoints field will contain only ULID strings for efficient storage.
        """
        # Start with existing datapoint IDs (these are already uploaded)
        if response.datapoints:
            datapoint_ids = response.datapoints.model_dump(mode="json")
        else:
            datapoint_ids = {}

        # Set ownership context for all datapoint uploads
        with ownership_context(
            workcell_id=self.workcell_definition.manager_id,
            workflow_id=wf.workflow_id,
            node_id=self.state_handler.get_node(step.node).info.node_id
            if self.state_handler.get_node(step.node).info
            else None,
            step_id=step.step_id,
        ):
            # Upload JSON result as ValueDataPoint if present
            if response.json_result is not None:
                datapoint = ValueDataPoint(
                    label="json_result",
                    value=response.json_result,
                )
                submitted_datapoint = self.data_client.submit_datapoint(datapoint)
                datapoint_ids["json_result"] = submitted_datapoint.datapoint_id
                self.logger.log_debug(
                    f"Uploaded JSON result as datapoint: {submitted_datapoint.datapoint_id}"
                )

            # Upload files as FileDataPoints if present
            if response.files:
                if isinstance(response.files, ActionFiles):
                    # Multiple files in ActionFiles object
                    response_files = response.files.model_dump(mode="json")
                    for file_key, file_path in response_files.items():
                        datapoint = FileDataPoint(
                            label=file_key,
                            path=str(file_path),
                        )
                        submitted_datapoint = self.data_client.submit_datapoint(
                            datapoint
                        )
                        datapoint_ids[file_key] = submitted_datapoint.datapoint_id
                        self.logger.log_debug(
                            f"Uploaded file '{file_key}' as datapoint: {submitted_datapoint.datapoint_id}"
                        )
                else:
                    # Single file Path
                    datapoint = FileDataPoint(
                        label="file",
                        path=str(response.files),
                    )
                    submitted_datapoint = self.data_client.submit_datapoint(datapoint)
                    datapoint_ids["file"] = submitted_datapoint.datapoint_id
                    self.logger.log_debug(
                        f"Uploaded file as datapoint: {submitted_datapoint.datapoint_id}"
                    )

            # Update response to contain only datapoint IDs
            response.datapoints = ActionDatapoints.model_validate(datapoint_ids)

            # Clear the original data now that it's stored as datapoints
            # This ensures we only store IDs in workflows for efficiency
            response.json_result = None
            response.files = None

            return response

    def update_active_nodes(
        self, state_manager: WorkcellStateHandler, update_info: bool = False
    ) -> None:
        """Update all active nodes in the workcell.

        Args:
            state_manager: The workcell state handler
            update_info: Whether to update node info in addition to status and state (default: False)
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            node_futures = []
            for node_name, node in state_manager.get_nodes().items():
                node_future = executor.submit(
                    self.update_node, node_name, node, state_manager, update_info
                )
                node_futures.append(node_future)

            # Wait for all node updates to complete
            concurrent.futures.wait(node_futures)

    def update_node(
        self,
        node_name: str,
        node: Node,
        state_manager: WorkcellStateHandler,
        update_info: bool = False,
    ) -> None:
        """Update a single node's status, state, and optionally info.

        Args:
            node_name: The name of the node to update
            node: The node object to update
            state_manager: The workcell state handler
            update_info: Whether to update node info (default: False). Node info changes
                        infrequently, so it's updated less often to reduce network overhead.
        """
        try:
            client = find_node_client(node.node_url)
            node.status = client.get_status()
            if update_info or node.info is None:
                node.info = client.get_info()
            node.state = client.get_state()
            with state_manager.wc_state_lock():
                state_manager.set_node(node_name, node)
        except Exception as e:
            error = Error.from_exception(e)
            node.status = NodeStatus(errored=True, errors=[error])
            with state_manager.wc_state_lock():
                state_manager.set_node(node_name, node)
            with ownership_context(
                workcell_id=self.workcell_definition.manager_id,
                node_id=node.info.node_id if node.info else None,
            ):
                self.logger.warning(
                    event=Event(
                        event_type=EventType.NODE_STATUS_UPDATE,
                        event_data=node.status,
                    )
                )



"""utility functions for the workcell"""

from madsci.client.node import NODE_CLIENT_MAP, AbstractNodeClient


def find_node_client(url: str) -> AbstractNodeClient:
    """Finds the appropriate node client based on a given node url"""
    for client in NODE_CLIENT_MAP.values():
        if client.validate_url(url):
            return client(url)
    for client in AbstractNodeClient.__subclasses__():
        if client.validate_url(url):
            return client(url)
    return None



"""Default MADSci Workcell scheduler"""

import traceback
from datetime import datetime, timedelta

from madsci.common.types.action_types import ActionStatus
from madsci.common.types.step_types import Step
from madsci.common.types.workflow_types import (
    SchedulerMetadata,
    Workflow,
)
from madsci.workcell_manager.condition_checks import evaluate_condition_checks
from madsci.workcell_manager.schedulers.scheduler import AbstractScheduler
from madsci.workcell_manager.workflow_utils import prepare_workflow_step


class Scheduler(AbstractScheduler):
    """
    This is the default scheduler for the MADSci Workcell Manager. It is a simple FIFO scheduler that checks if the workflow is ready to run.

    - It checks a variety of conditions to determine if a workflow is ready to run. If the workflow is not ready to run, it will add a reason to the scheduler metadata for the workflow.
    - It sets the priority of the workflow based on the order in which the workflows were submitted.
    """

    def run_iteration(self, workflows: list[Workflow]) -> dict[str, SchedulerMetadata]:
        """Run an iteration of the scheduling algorithm and return a mapping of workflow IDs to SchedulerMetadata"""
        priority = 0
        workflows = sorted(
            workflows,
            key=lambda item: item.submitted_time,
        )
        workflow_definition_metadata_map = {}

        for wf in workflows:
            try:
                metadata = wf.scheduler_metadata
                metadata.ready_to_run = True
                metadata.reasons = []

                if wf.status.current_step_index < len(wf.steps):
                    step = wf.steps[wf.status.current_step_index]
                    updated_step = prepare_workflow_step(
                        self.workcell_definition,
                        self.state_handler,
                        step,
                        wf,
                        location_client=self.location_client,
                    )
                    self.check_workflow_status(wf, metadata)
                    self.location_checks(updated_step, metadata)
                    self.resource_checks(updated_step, metadata)
                    self.node_checks(updated_step, wf, metadata)
                    self.step_checks(updated_step, metadata)
                    metadata = evaluate_condition_checks(updated_step, self, metadata)
                    metadata.priority = priority
                    priority -= 1

            except Exception as e:
                self.logger.error(
                    f"Error in scheduler while evaluating workflow {wf.workflow_id}: {traceback.format_exc()}"
                )
                metadata.ready_to_run = False
                metadata.reasons.append(f"Exception in scheduler: {e}")
            finally:
                workflow_definition_metadata_map[wf.workflow_id] = metadata

        return workflow_definition_metadata_map

    def check_workflow_status(self, wf: Workflow, metadata: SchedulerMetadata) -> None:
        """Check if the workflow is ready to run (i.e. not paused, not completed, etc.)"""
        if wf.status.paused:
            metadata.ready_to_run = False
            metadata.reasons.append("Workflow is paused")
            return
        if wf.status.running:
            metadata.ready_to_run = False
            metadata.reasons.append("Workflow is already running")
            return
        if not wf.status.active:
            metadata.ready_to_run = False
            metadata.reasons.append(
                f"Workflow must be active (i.e., not completed, failed, cancelled, or paused), not {wf.status.description}"
            )

    def location_checks(self, step: Step, metadata: SchedulerMetadata) -> None:
        """Check if the location(s) for the step are ready"""
        for location in step.locations.values():
            if location is None:
                continue
            # Get the current location state from LocationManager
            current_location = location
            try:
                if self.location_client is not None:
                    current_location = self.location_client.get_location(
                        location.location_id
                    )
            except Exception:
                # If LocationManager is not available or location not found, use step's location
                current_location = location

            if (
                current_location.resource_id is not None
                and self.resource_client is not None
            ):
                self.resource_client.get_resource(current_location.resource_id)
                # TODO: what do we do with the location_resource?
            if current_location.reservation is not None:
                metadata.ready_to_run = False
                metadata.reasons.append(
                    f"Location {current_location.location_id} is reserved by {current_location.reservation.owned_by.model_dump(mode='json', exclude_none=True)}"
                )

    def resource_checks(self, step: Step, metadata: SchedulerMetadata) -> None:
        """Check if the resources for the step are ready TODO: actually check"""

    def step_checks(self, step: Step, metadata: SchedulerMetadata) -> None:
        """Check if the step was not ready recently mark the workflow as not active"""
        if (
            step.result is not None
            and step.result.status == ActionStatus.NOT_READY
            and datetime.now().astimezone() - step.result.history_created_at
            < timedelta(seconds=30)
        ):
            metadata.ready_to_run = False
            metadata.reasons.append(f"Waiting for Node {step.node} to be ready")

    def node_checks(
        self, step: Step, wf: Workflow, metadata: SchedulerMetadata
    ) -> None:
        """Check if the node used in the step currently has a "ready" status"""
        if step.node is not None:
            node = self.state_handler.get_node(step.node)
            if node is None:
                metadata.ready_to_run = False
                metadata.reasons.append(f"Node {step.node} not found")
            if not node.status.ready:
                metadata.ready_to_run = False
                metadata.reasons.append(
                    f"Node {step.node} not ready: {node.status.description}"
                )
            # Check if node is locked (another workflow is currently sending it an action request)
            node_lock = self.state_handler.node_lock(step.node)
            if node_lock.locked():
                metadata.ready_to_run = False
                metadata.reasons.append(f"Node {step.node} is locked by another action")

            if node.reservation is not None and node.reservation.check(
                wf.ownership_info
            ):
                metadata.ready_to_run = False
                metadata.reasons.append(
                    f"Node {step.node} is reserved by {node.reservation.owned_by.model_dump(mode='json', exclude_none=True)}"
                )



"""Pytest unit tests for the MADSci Workcell Manager MongoDB migration tools."""

import json
import os
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from madsci.common.mongodb_migration_tool import (
    MongoDBMigrationSettings,
    MongoDBMigrator,
    main,
)
from madsci.common.mongodb_version_checker import MongoDBVersionChecker
from pydantic_extra_types.semantic_version import SemanticVersion


@pytest.fixture
def temp_workcell_schema():
    """Create temporary workcell manager schema for testing"""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Create workcell manager schema
        workcell_manager_dir = temp_path / "madsci" / "workcell_manager"
        workcell_manager_dir.mkdir(parents=True)
        workcell_schema = {
            "database": "madsci_workcells",
            "schema_version": "1.0.0",
            "description": "Schema definition for MADSci Workcell Manager MongoDB",
            "collections": {
                "archived_workflows": {
                    "description": "Main archived workflows collection",
                    "indexes": [],
                },
                "schema_versions": {
                    "description": "Version tracking",
                    "indexes": [
                        {
                            "keys": [["version", 1]],
                            "name": "version_unique",
                            "unique": True,
                        }
                    ],
                },
            },
        }
        (workcell_manager_dir / "schema.json").write_text(json.dumps(workcell_schema))

        # Change to temp directory
        original_cwd = Path.cwd()
        os.chdir(temp_path)

        try:
            yield workcell_manager_dir / "schema.json"
        finally:
            os.chdir(original_cwd)


@pytest.fixture
def mock_mongo_client_workcells():
    """Mock MongoDB client for workcells database"""
    with patch("madsci.common.mongodb_migration_tool.MongoClient") as mock_client:
        mock_db = Mock()
        mock_collection = Mock()

        # Setup mock for workcells database - fix the dictionary access
        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        mock_db.list_collection_names.return_value = ["archived_workflows"]
        mock_db.__getitem__ = Mock(return_value=mock_collection)
        mock_db.create_collection = Mock()

        yield mock_client


def test_workcell_version_checker_detects_missing_version_tracking():
    """Test that version checker detects workcells database without version tracking"""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_db.list_collection_names.return_value = [
            "archived_workflows"
        ]  # No schema_versions

        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        schema_file = Path("workcell_schema.json")
        schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

        try:
            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "madsci_workcells", str(schema_file)
            )

            db_version = checker.get_database_version()
            assert db_version == SemanticVersion(0, 0, 0)

            needs_migration, _, _ = checker.is_migration_needed()
            assert needs_migration is True
        finally:
            schema_file.unlink()


def test_workcell_version_checker_auto_initializes_fresh_database():
    """Test that version checker auto-initializes completely fresh databases"""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_collection = Mock()

        # Mock a completely fresh database (no collections)
        mock_db.list_collection_names.return_value = []

        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        # Mock the schema_versions collection
        mock_db.__getitem__ = Mock(return_value=mock_collection)

        # Mock find_one to return None (no existing version record)
        mock_collection.find_one.return_value = None

        schema_file = Path("workcell_schema.json")
        schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

        try:
            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "madsci_workcells", str(schema_file)
            )

            # This should auto-initialize without raising an error
            checker.validate_or_fail()

            # Verify that create_schema_versions_collection was called
            mock_collection.create_index.assert_called()
            mock_collection.insert_one.assert_called()
        finally:
            schema_file.unlink()


def test_workcell_version_checker_still_requires_migration_for_existing_db():
    """Test that version checker still requires manual migration for existing databases without version tracking"""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_db.list_collection_names.return_value = [
            "archived_workflows"
        ]  # Existing collections but no schema_versions

        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        schema_file = Path("workcell_schema.json")
        schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

        try:
            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "madsci_workcells", str(schema_file)
            )

            # This should still raise an error requiring manual migration
            with pytest.raises(
                RuntimeError, match="needs version tracking initialization"
            ):
                checker.validate_or_fail()
        finally:
            schema_file.unlink()


@patch("madsci.common.mongodb_migration_tool.MongoDBMigrator")
@patch(
    "sys.argv",
    [
        "migration_tool.py",
        "--database",
        "madsci_workcells",
        "--backup_only",
        "true",
    ],
)
def test_workcell_backup_only_command(mock_migrator_class):
    """Test backup only command for workcells database"""
    mock_migrator = Mock()
    backup_dir = Path(tempfile.gettempdir()) / "madsci_workcells_backup"
    mock_migrator.backup_tool.create_backup.return_value = backup_dir
    mock_migrator_class.return_value = mock_migrator

    with patch.dict(os.environ, {"MONGODB_URL": "mongodb://localhost:27017"}):
        # Create temp schema structure for auto-detection
        temp_dir = Path(tempfile.gettempdir()) / "test_backup_command"
        temp_dir.mkdir(exist_ok=True)
        schema_dir = temp_dir / "madsci" / "workcell_manager"
        schema_dir.mkdir(parents=True, exist_ok=True)
        temp_schema = schema_dir / "schema.json"
        temp_schema.write_text(json.dumps({"schema_version": "1.0.0"}))

        original_cwd = Path.cwd()
        os.chdir(temp_dir)

        try:
            main()
        finally:
            os.chdir(original_cwd)

    # Verify only backup was called
    mock_migrator.backup_tool.create_backup.assert_called_once()
    mock_migrator.run_migration.assert_not_called()


@patch("subprocess.run")
def test_workcell_backup_creation(mock_subprocess):
    """Test workcells database backup creation using mongodump"""

    def mock_mongodump(*args, **_kwargs):
        """Mock mongodump by creating expected directory structure."""
        # Extract backup path from mongodump command
        cmd = args[0]
        out_index = cmd.index("--out") + 1
        backup_path = Path(cmd[out_index])
        db_name = cmd[cmd.index("--db") + 1]

        # Create the directory structure that mongodump would create
        db_backup_path = backup_path / db_name
        db_backup_path.mkdir(parents=True, exist_ok=True)

        # Create a mock collection file
        (db_backup_path / "workcells.bson").touch()
        (db_backup_path / "workcells.metadata.json").touch()

        mock_result = Mock()
        mock_result.returncode = 0
        mock_result.stderr = ""
        return mock_result

    mock_subprocess.side_effect = mock_mongodump

    schema_file = Path("workcell_schema.json")
    schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

    try:
        settings = MongoDBMigrationSettings(
            mongo_db_url="mongodb://localhost:27017",
            database="madsci_workcells",
            schema_file=str(schema_file),
        )
        migrator = MongoDBMigrator(settings)

        # Mock the post-backup processing to bypass filesystem checks
        with patch.object(migrator.backup_tool, "_post_backup_processing"):
            backup_path = migrator.backup_tool.create_backup()

            # Verify backup path format
            assert "madsci_workcells_backup_" in backup_path.name

            # Verify mongodump was called with workcells database
            mock_subprocess.assert_called_once()
            call_args = mock_subprocess.call_args[0][0]
            assert "mongodump" in call_args
            assert "--db" in call_args
            assert "madsci_workcells" in call_args
    finally:
        schema_file.unlink()


def test_workcell_collection_creation(mock_mongo_client_workcells):  # noqa
    """Test archived workflows collection creation during migration"""
    schema_file = Path("workcell_schema.json")
    schema_content = {
        "schema_version": "1.0.0",
        "collections": {
            "archived_workflows": {"indexes": []},
            "schema_versions": {"indexes": []},
        },
    }
    schema_file.write_text(json.dumps(schema_content))

    try:
        settings = MongoDBMigrationSettings(
            mongo_db_url="mongodb://localhost:27017",
            database="madsci_workcells",
            schema_file=str(schema_file),
        )
        migrator = MongoDBMigrator(settings)

        # Access the database from the migrator instance, not the mock client
        migrator.database.list_collection_names.return_value = []

        migrator._ensure_collection_exists("archived_workflows")

        # Verify archived workflows collection creation was called
        migrator.database.create_collection.assert_called_with("archived_workflows")
    finally:
        schema_file.unlink()


@patch("madsci.common.mongodb_migration_tool.MongoDBMigrator")
@patch(
    "sys.argv",
    [
        "migration_tool.py",
        "--database",
        "madsci_workcells",
        "--check_version",
        "true",
    ],
)
def test_workcell_check_version_command(mock_migrator_class):
    """Test check version command for workcells database"""
    mock_migrator = Mock()
    mock_migrator_class.return_value = mock_migrator

    # Mock version checker
    mock_version_checker = Mock()
    mock_version_checker.is_migration_needed.return_value = (False, "1.0.0", "1.0.0")
    mock_migrator.version_checker = mock_version_checker

    with patch.dict(os.environ, {"MONGODB_URL": "mongodb://localhost:27017"}):
        # Create temp schema structure for auto-detection
        temp_dir = Path(tempfile.gettempdir()) / "test_check_version"
        temp_dir.mkdir(exist_ok=True)
        schema_dir = temp_dir / "madsci" / "workcell_manager"
        schema_dir.mkdir(parents=True, exist_ok=True)
        temp_schema = schema_dir / "schema.json"
        temp_schema.write_text(json.dumps({"schema_version": "1.0.0"}))

        original_cwd = Path.cwd()
        os.chdir(temp_dir)

        try:
            main()
        finally:
            os.chdir(original_cwd)

    # Verify version check was called
    mock_version_checker.is_migration_needed.assert_called_once()
    mock_migrator.run_migration.assert_not_called()


def test_workcell_version_mismatch_detection():
    """Test detection of version mismatches in workcells database (major/minor versions)"""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_collection = Mock()

        # Mock existing version tracking with different major/minor version
        mock_collection.find_one.return_value = {"version": "0.9.0"}
        mock_db.list_collection_names.return_value = [
            "archived_workflows",
            "schema_versions",
        ]

        # Fix the dictionary access
        mock_db.__getitem__ = Mock(return_value=mock_collection)
        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        schema_file = Path("workcell_schema.json")
        schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

        try:
            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "madsci_workcells", str(schema_file)
            )

            # Schema version is 1.0.0 (minor version difference from DB 0.9.0)
            needs_migration, _, db_version = checker.is_migration_needed()
            assert needs_migration is True
            assert db_version == "0.9.0"
        finally:
            schema_file.unlink()


def test_workcell_patch_version_differences():
    """Test that patch version differences trigger migration in workcells database (schema versions must match exactly)"""
    with patch("madsci.common.mongodb_version_checker.MongoClient") as mock_client:
        mock_db = Mock()
        mock_collection = Mock()

        # Mock existing version tracking with same major.minor but different patch
        mock_collection.find_one.return_value = {"version": "1.0.0"}
        mock_db.list_collection_names.return_value = [
            "archived_workflows",
            "schema_versions",
        ]

        # Fix the dictionary access
        mock_db.__getitem__ = Mock(return_value=mock_collection)
        mock_client_instance = Mock()
        mock_client_instance.__getitem__ = Mock(return_value=mock_db)
        mock_client.return_value = mock_client_instance

        schema_file = Path("workcell_schema.json")
        schema_file.write_text(json.dumps({"schema_version": "1.0.1"}))

        try:
            checker = MongoDBVersionChecker(
                "mongodb://localhost:27017", "madsci_workcells", str(schema_file)
            )

            # Schema version is 1.0.1 (patch version difference from DB 1.0.0)
            needs_migration, _, db_version = checker.is_migration_needed()
            assert needs_migration is True  # Should need migration since 1.0.1 != 1.0.0
            assert db_version == SemanticVersion.parse("1.0.0")
        finally:
            schema_file.unlink()


@patch("madsci.common.mongodb_migration_tool.MongoDBMigrator")
@patch(
    "sys.argv",
    [
        "migration_tool.py",
        "--database",
        "madsci_workcells",
        "--restore_from",
        "workcells_backup",
    ],
)
def test_workcell_restore_command(mock_migrator_class):
    """Test restore command for workcells database"""
    mock_migrator = Mock()
    mock_migrator_class.return_value = mock_migrator

    with patch.dict(os.environ, {"MONGODB_URL": "mongodb://localhost:27017"}):
        # Create temp schema structure for auto-detection
        temp_dir = Path(tempfile.gettempdir()) / "test_restore"
        temp_dir.mkdir(exist_ok=True)
        schema_dir = temp_dir / "madsci" / "workcell_manager"
        schema_dir.mkdir(parents=True, exist_ok=True)
        temp_schema = schema_dir / "schema.json"
        temp_schema.write_text(json.dumps({"schema_version": "1.0.0"}))

        original_cwd = Path.cwd()
        os.chdir(temp_dir)

        try:
            main()
        finally:
            os.chdir(original_cwd)

    # Verify restore was called with correct path
    mock_migrator.backup_tool.restore_from_backup.assert_called_once_with(
        Path("workcells_backup")
    )
    mock_migrator.run_migration.assert_not_called()


def test_workcell_schema_file_detection():
    """Test automatic detection of workcell manager schema file"""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Create workcell manager schema directory
        workcell_manager_dir = temp_path / "madsci" / "workcell_manager"
        workcell_manager_dir.mkdir(parents=True)
        schema_file = workcell_manager_dir / "schema.json"
        schema_file.write_text(json.dumps({"schema_version": "1.0.0"}))

        original_cwd = Path.cwd()
        os.chdir(temp_path)

        try:
            settings = MongoDBMigrationSettings(database="madsci_workcells")
            detected_path = settings.get_effective_schema_file_path()
            assert detected_path.name == "schema.json"
            assert "workcell_manager" in str(detected_path)
        finally:
            os.chdir(original_cwd)



"""MADSci Workcell Manager using AbstractManagerBase."""

import json
import traceback
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Annotated, Any, AsyncGenerator, ClassVar, Optional, Union

from classy_fastapi import get, post
from fastapi import FastAPI, Form, HTTPException, UploadFile
from fastapi.params import Body
from madsci.client.data_client import DataClient
from madsci.client.location_client import (
    LocationClient,
)
from madsci.common.context import (
    get_current_madsci_context,
)
from madsci.common.manager_base import AbstractManagerBase
from madsci.common.mongodb_version_checker import MongoDBVersionChecker
from madsci.common.ownership import global_ownership_info, ownership_context
from madsci.common.types.admin_command_types import AdminCommandResponse
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.event_types import Event, EventType
from madsci.common.types.mongodb_migration_types import MongoDBMigrationSettings
from madsci.common.types.node_types import Node
from madsci.common.types.workcell_types import (
    WorkcellManagerDefinition,
    WorkcellManagerHealth,
    WorkcellManagerSettings,
    WorkcellState,
)
from madsci.common.types.workflow_types import (
    Workflow,
    WorkflowDefinition,
)
from madsci.workcell_manager.state_handler import WorkcellStateHandler
from madsci.workcell_manager.workcell_engine import Engine
from madsci.workcell_manager.workcell_utils import find_node_client
from madsci.workcell_manager.workflow_utils import (
    check_parameters,
    create_workflow,
    save_workflow_files,
)
from pymongo.synchronous.database import Database
from ulid import ULID

# Module-level constants for Body() calls to avoid B008 linting errors
LOOKUP_VAL_BODY = Body(...)


class WorkcellManager(
    AbstractManagerBase[WorkcellManagerSettings, WorkcellManagerDefinition]
):
    """
    MADSci Workcell Manager using the new AbstractManagerBase pattern.

    This manager uses MadsciClientMixin (via AbstractManagerBase) for client management.
    Required clients: data, location
    """

    SETTINGS_CLASS = WorkcellManagerSettings
    DEFINITION_CLASS = WorkcellManagerDefinition

    # Declare required clients for the mixin
    REQUIRED_CLIENTS: ClassVar[list[str]] = ["event", "data", "location"]

    def __init__(
        self,
        settings: Optional[WorkcellManagerSettings] = None,
        definition: Optional[WorkcellManagerDefinition] = None,
        redis_connection: Optional[Any] = None,
        mongo_connection: Optional[Database] = None,
        start_engine: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initialize the WorkcellManager."""
        self.redis_connection = redis_connection
        self.mongo_connection = mongo_connection
        self.start_engine = start_engine

        super().__init__(settings=settings, definition=definition, **kwargs)

    def create_default_definition(self) -> WorkcellManagerDefinition:
        """Create a default definition instance for this manager."""
        name = str(self.get_definition_path().name).split(".")[0]
        return WorkcellManagerDefinition(name=name)

    def initialize(self, **kwargs: Any) -> None:
        """
        Initialize manager-specific components.

        This method sets up the workcell-specific state handler and clients.
        Client initialization is handled by MadsciClientMixin via setup_clients().
        """
        super().initialize(**kwargs)

        # Skip version validation if external mongo_connection was provided (e.g., in tests)
        # This is commonly done in tests where a mock or containerized MongoDB is used
        if self.mongo_connection is not None:
            # External connection provided, likely in test context - skip version validation
            self.logger.info(
                "External mongo_connection provided, skipping MongoDB version validation"
            )
            # Continue with the rest of initialization (ownership, state handler, clients)
            global_ownership_info.workcell_id = self.definition.manager_id
            global_ownership_info.manager_id = self.definition.manager_id

            # Initialize state handler
            self.state_handler = WorkcellStateHandler(
                self.definition,
                workcell_settings=self.settings,
                redis_connection=self.redis_connection,
                mongo_connection=self.mongo_connection,
            )

            # Initialize clients
            context = get_current_madsci_context()
            self.data_client = DataClient(context.data_server_url)
            self.location_client = LocationClient(context.location_server_url)
            return

        self.logger.info("Validating MongoDB schema version...")

        schema_file_path = Path(__file__).parent / "schema.json"
        mig_cfg = MongoDBMigrationSettings(database=self.settings.database_name)
        version_checker = MongoDBVersionChecker(
            db_url=str(self.settings.mongo_db_url),
            database_name=self.settings.database_name,
            schema_file_path=str(schema_file_path),
            backup_dir=str(mig_cfg.backup_dir),
            logger=self.logger,
        )

        try:
            version_checker.validate_or_fail()
            self.logger.info("MongoDB version validation completed successfully")
        except RuntimeError as e:
            self.logger.error(
                "DATABASE VERSION MISMATCH DETECTED! SERVER STARTUP ABORTED!"
            )
            raise e

        # Set up global ownership
        global_ownership_info.workcell_id = self.definition.manager_id
        global_ownership_info.manager_id = self.definition.manager_id

        # Initialize state handler
        self.state_handler = WorkcellStateHandler(
            self.definition,
            workcell_settings=self.settings,
            redis_connection=self.redis_connection,
            mongo_connection=self.mongo_connection,
        )

        # Initialize clients using MadsciClientMixin
        # This will create data_client and location_client from context
        self.setup_clients()

        # Clients are now available as self.data_client and self.location_client
        # They will use URLs from get_current_madsci_context() by default

    def create_server(self, **kwargs: Any) -> FastAPI:
        """Create the FastAPI server application with lifespan."""

        # Set up lifespan context manager
        @asynccontextmanager
        async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
            """Start the REST server and initialize the state handler and engine"""
            _ = app  # Mark app as used to avoid lint warning
            global_ownership_info.workcell_id = self.definition.manager_id
            global_ownership_info.manager_id = self.definition.manager_id

            # LOG WORKCELL START EVENT
            self.logger.log(
                Event(
                    event_type=EventType.WORKCELL_START,
                    event_data=self.definition.model_dump(mode="json"),
                )
            )

            if self.start_engine:
                engine = Engine(self.state_handler, self.data_client)
                engine.spin()
            else:
                with self.state_handler.wc_state_lock():
                    self.state_handler.initialize_workcell_state()
            try:
                yield
            finally:
                # LOG WORKCELL STOP EVENT
                self.logger.log(
                    Event(
                        event_type=EventType.WORKCELL_STOP,
                        event_data=self.definition.model_dump(mode="json"),
                    )
                )

        # Create app with lifespan
        app = FastAPI(
            title=self._definition.name,
            description=self._definition.description
            or f"{self._definition.name} Manager",
            lifespan=lifespan,
            **kwargs,
        )

        # Configure the app (middleware, etc.)
        self.configure_app(app)

        # Include the router from this Routable class
        app.include_router(self.router)

        return app

    # Endpoint implementations

    def get_health(self) -> WorkcellManagerHealth:
        """Get the health status of the Workcell Manager."""
        health = WorkcellManagerHealth()

        try:
            # Test Redis connection if configured
            if (
                hasattr(self.state_handler, "_redis_client")
                and self.state_handler._redis_client
            ):
                self.state_handler._redis_client.ping()
                health.redis_connected = True
            else:
                health.redis_connected = None

            # Count nodes and check their reachability
            total_nodes = len(self.definition.nodes)
            health.total_nodes = total_nodes

            # TODO: Implement actual node reachability checks
            health.nodes_reachable = total_nodes

            health.healthy = True
            health.description = "Workcell Manager is running normally"

        except Exception as e:
            health.healthy = False
            if "redis" in str(e).lower():
                health.redis_connected = False
            health.description = f"Health check failed: {e!s}"

        return health

    @get("/workcell")
    def get_workcell(self) -> WorkcellManagerDefinition:
        """Get the currently running workcell (backward compatibility)."""
        return self.state_handler.get_workcell_definition()

    @get("/state")
    def get_state(self) -> WorkcellState:
        """Get the current state of the workcell."""
        return self.state_handler.get_workcell_state()

    @get("/nodes")
    def get_nodes(self) -> dict[str, Node]:
        """Get info on the nodes in the workcell."""
        return self.state_handler.get_nodes()

    @get("/node/{node_name}")
    def get_node(self, node_name: str) -> Union[Node, str]:
        """Get information about about a specific node."""
        try:
            node = self.state_handler.get_node(node_name)
        except Exception:
            return "Node not found!"
        return node

    @post("/node")
    def add_node(
        self,
        node_name: str,
        node_url: str,
        permanent: bool = False,
    ) -> Union[Node, str]:
        """Add a node to the workcell's node list"""
        if node_name in self.state_handler.get_nodes():
            return "Node name exists, node names must be unique!"
        node = Node(node_url=node_url)
        self.state_handler.set_node(node_name, node)
        if permanent:
            workcell = self.state_handler.get_workcell_definition()
            workcell.nodes[node_name] = node_url
            workcell_path = self.get_definition_path()
            if workcell_path.exists():
                workcell.to_yaml(workcell_path)
            self.state_handler.set_workcell_definition(workcell)

        return self.state_handler.get_node(node_name)

    @post("/admin/{command}")
    def send_admin_command(self, command: str) -> list:
        """Send an admin command to all capable nodes."""
        responses = []
        for node in self.state_handler.get_nodes().values():
            if command in node.info.capabilities.admin_commands:
                client = find_node_client(node.node_url)
                response = client.send_admin_command(command)
                responses.append(response)
        return responses

    @post("/admin/{command}/{node}")
    def send_admin_command_to_node(
        self, command: str, node: str
    ) -> AdminCommandResponse:
        """Send admin command to a node."""
        node = self.state_handler.get_node(node)
        if command in node.info.capabilities.admin_commands:
            client = find_node_client(node.node_url)
            return client.send_admin_command(command)
        raise HTTPException(
            status_code=400, detail="Node cannot perform that admin command"
        )

    @get("/workflows/active")
    def get_active_workflows(self) -> dict[str, Workflow]:
        """Get active workflows."""
        return self.state_handler.get_active_workflows()

    @get("/workflows/archived")
    def get_archived_workflows(self, number: int = 20) -> dict[str, Workflow]:
        """Get archived workflows."""
        return self.state_handler.get_archived_workflows(number)

    @get("/workflows/queue")
    def get_workflow_queue(self) -> list[Workflow]:
        """Get all queued workflows."""
        return self.state_handler.get_workflow_queue()

    @get("/workflow/{workflow_id}")
    def get_workflow(self, workflow_id: str) -> Workflow:
        """Get info on a specific workflow."""
        return self.state_handler.get_workflow(workflow_id)

    @post("/workflow/{workflow_id}/pause")
    def pause_workflow(self, workflow_id: str) -> Workflow:
        """Pause a specific workflow."""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_active_workflow(workflow_id)
            if wf.status.active:
                if wf.status.running:
                    self.send_admin_command_to_node(
                        "pause", wf.steps[wf.status.current_step_index].node
                    )
                wf.status.paused = True
                self.state_handler.set_active_workflow(wf)

        return self.state_handler.get_active_workflow(workflow_id)

    @post("/workflow/{workflow_id}/resume")
    def resume_workflow(self, workflow_id: str) -> Workflow:
        """Resume a paused workflow."""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_active_workflow(workflow_id)
            if wf.status.paused:
                if wf.status.running:
                    self.send_admin_command_to_node(
                        "resume", wf.steps[wf.status.current_step_index].node
                    )
                wf.status.paused = False
                self.state_handler.set_active_workflow(wf)
                self.state_handler.enqueue_workflow(wf.workflow_id)
        return self.state_handler.get_active_workflow(workflow_id)

    @post("/workflow/{workflow_id}/cancel")
    def cancel_workflow(self, workflow_id: str) -> Workflow:
        """Cancel a specific workflow."""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_workflow(workflow_id)
            if wf.status.running:
                self.send_admin_command_to_node(
                    "cancel", wf.steps[wf.status.current_step_index].node
                )
            wf.status.cancelled = True
            self.state_handler.set_active_workflow(wf)
        return self.state_handler.get_active_workflow(workflow_id)

    @post("/workflow/{workflow_id}/retry")
    def retry_workflow(self, workflow_id: str, index: int = -1) -> Workflow:
        """Retry an existing workflow from a specific step."""
        with self.state_handler.wc_state_lock():
            wf = self.state_handler.get_workflow(workflow_id)
            if wf.status.terminal:
                if index < 0:
                    index = wf.status.current_step_index
                wf.status.reset(index)
                self.state_handler.set_active_workflow(wf)
                self.state_handler.delete_archived_workflow(wf.workflow_id)
                self.state_handler.enqueue_workflow(wf.workflow_id)
            else:
                raise HTTPException(
                    status_code=400,
                    detail="Workflow is not in a terminal state, cannot retry",
                )
        return self.state_handler.get_active_workflow(workflow_id)

    @post("/workflow_definition")
    async def submit_workflow_definition(
        self,
        workflow_definition: WorkflowDefinition,
    ) -> str:
        """
        Parses the payload and workflow files, and then pushes a workflow job onto the redis queue

        Parameters
        ----------
        workflow_definition: YAML string
        - The workflow_definition yaml file


        Returns
        -------
        response: Workflow Definition ID
        - the workflow definition ID
        """
        try:
            try:
                wf_def = WorkflowDefinition.model_validate(workflow_definition)

            except Exception as e:
                traceback.print_exc()
                raise HTTPException(status_code=422, detail=str(e)) from e
            return self.state_handler.save_workflow_definition(
                workflow_definition=wf_def,
            )
        except HTTPException as e:
            raise e
        except Exception as e:
            self.logger.error(f"Error saving workflow definition: {e}")
            traceback.print_exc()
            raise HTTPException(
                status_code=500,
                detail=f"Error saving workflow definition: {e}",
            ) from e

    @get("/workflow_definition/{workflow_definition_id}")
    async def get_workflow_definition(
        self,
        workflow_definition_id: str,
    ) -> WorkflowDefinition:
        """
        Parses the payload and workflow files, and then pushes a workflow job onto the redis queue

        Parameters
        ----------
        Workflow Definition ID: str
        - the workflow definition ID

        Returns
        -------
        response: WorkflowDefinition
        - a workflow run object for the requested run_id
        """
        try:
            return self.state_handler.get_workflow_definition(workflow_definition_id)
        except Exception as e:
            traceback.print_exc()
            raise HTTPException(status_code=404, detail=str(e)) from e

    @post("/workflow")
    async def start_workflow(
        self,
        workflow_definition_id: Annotated[str, Form()],
        ownership_info: Annotated[Optional[str], Form()] = None,
        json_inputs: Annotated[Optional[str], Form()] = None,
        file_input_paths: Annotated[Optional[str], Form()] = None,
        files: list[UploadFile] = [],
    ) -> Workflow:
        """
        Parses the payload and workflow files, and then pushes a workflow job onto the redis queue

        Parameters
        ----------
        workflow: YAML string
        - The workflow yaml file
        parameters: Optional[Dict[str, Any]] = {}
        - Dynamic values to insert into the workflow file
        ownership_info: Optional[OwnershipInfo]
        - Information about the experiments, users, etc. that own this workflow
        simulate: bool
        - whether to use real robots or not
        validate_only: bool
        - whether to validate the workflow without queueing it

        Returns
        -------
        response: Workflow
        - a workflow run object for the requested run_id
        """
        try:
            try:
                workflow_id = ULID.from_str(workflow_definition_id)
                wf_def = self.state_handler.get_workflow_definition(str(workflow_id))

            except Exception as e:
                traceback.print_exc()
                raise HTTPException(status_code=422, detail=str(e)) from e

            ownership_info = (
                OwnershipInfo.model_validate_json(ownership_info)
                if ownership_info
                else OwnershipInfo()
            )
            with ownership_context(**ownership_info.model_dump(exclude_none=True)):
                if json_inputs is None or json_inputs == "":
                    json_inputs = {}
                else:
                    json_inputs = json.loads(json_inputs)
                    if not isinstance(json_inputs, dict) or not all(
                        isinstance(k, str) for k in json_inputs
                    ):
                        raise HTTPException(
                            status_code=400,
                            detail="Parameters must be a dictionary with string keys",
                        )
                if file_input_paths is None or file_input_paths == "":
                    file_input_paths = {}
                else:
                    file_input_paths = json.loads(file_input_paths)
                    if not isinstance(file_input_paths, dict) or not all(
                        isinstance(k, str) for k in file_input_paths
                    ):
                        raise HTTPException(
                            status_code=400,
                            detail="Input File Paths must be a dictionary with string keys",
                        )
                workcell = self.state_handler.get_workcell_definition()
                check_parameters(wf_def, json_inputs, file_input_paths)
                wf = create_workflow(
                    workflow_def=wf_def,
                    workcell=workcell,
                    json_inputs=json_inputs,
                    file_input_paths=file_input_paths,
                    state_handler=self.state_handler,
                    location_client=self.location_client,
                )

                wf = save_workflow_files(
                    workflow=wf, files=files, data_client=self.data_client
                )

                with self.state_handler.wc_state_lock():
                    self.state_handler.set_active_workflow(wf)
                    self.state_handler.enqueue_workflow(wf.workflow_id)

                self.logger.log(
                    Event(
                        event_type=EventType.WORKFLOW_START,
                        event_data=wf.model_dump(mode="json"),
                    )
                )
                return wf

        except HTTPException as e:
            raise e
        except Exception as e:
            self.logger.error(f"Error starting workflow: {e}")
            traceback.print_exc()
            raise HTTPException(
                status_code=500,
                detail=f"Error starting workflow: {e}",
            ) from e


if __name__ == "__main__":
    manager = WorkcellManager()
    manager.run_server()
